
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <string.h>

#include "Rtypes.h"
#include "TROOT.h"
#include "TFile.h"
#include "TRandom.h"
#include "TH1.h"
#include "TH2.h"
#include "TObjArray.h"
#include "TObject.h"
#include "TKey.h"
#include "TSystem.h"
#include "TCutG.h"
#include "TTree.h"
#include <assert.h>

#include "ctk.h"

#define GGMAX 10000
#define MAXL 500

#define DDEBUG 0

extern struct ANAPAR Pars;
extern float rad2deg;

TList *wlist;
TFile *f1;
TH1D *fmsp;
TH1D *hitpat;
TH1D *sumTrackE;
TH1D *sumehi;
TH1D *polAngle;
TH1D *dopfac;
TH1D *radius;
TH1D *radius_c[MAXDETNO];
TH1D *dcx2vsradius;
TH1D *dcx2vsradius_norm;
TH1D *dcx2;
TH1D *FPGAE;
TH1D *EXTtac;
TH1D *BGSEsi;
TH1D *sEXT_tac;
TH1D *decompSumE;
TH1D *G4_obs;
TH1D *G4_emit;
TH1D *ehi[MAXDETNO];
TH2F *gg;
TH2F *gg_decomp;
TH2F *gg_fpga;
TH2F *rad_e;
TH2F *SMAP_allhits;
TH2F *SMAP_firsthits;
TH2F *fomXe;
TH2F *S800_PID;

TCutG *PIDwin;

float egamBinWidth;
long nPID, nPIDwin, nGam;
extern double mean_pol, mean_azi;
extern int mean_n;


/*-----------------------------------------------------------*/

float
findPolarFromCartesian (float xx, float yy, float zz, float *rr)
{
  float d1;

  *rr = sqrtf (xx * xx + yy * yy + zz * zz);
  d1 = acosf (zz / *rr);

  return (d1);
}

/*-----------------------------------------------------------*/

float
findAzimuthFromCartesian (float xx, float yy, float zz)
{

  float d1;

#if(0)
  if (xx > 0 && yy >= 0)
    d1 = atanf (yy / xx);
  else if (xx > 0 && yy < 0)
    d1 = atanf (yy / xx) + 2. * M_PI;
  else if (xx < 0)
    d1 = atanf (yy / xx) + M_PI;
  else if (xx == 0 && yy > 0)
    d1 = M_PI / 2.;
  else if (xx == 0 && yy < 0)
    d1 = 3. * M_PI / 2.;
  else
    d1 = 0.0;
#endif
  d1 = atan2f (yy, xx);


  return (d1);

}

/*-----------------------------------------------------------*/

TCutG *
rd2dwin (Char_t * winname)
{

  /* declarations */

  char str[STRLEN];
  TCutG *mycutg;

  TFile *f = new TFile (winname, "read");
  mycutg = (TCutG *) f->Get (winname);

  if (mycutg != NULL)
    {
      printf ("2dwin read from file:\n");
      sprintf (str, "ls -l %s", winname);
      gSystem->Exec (str);
      /* mycutg->Print(); */
      fflush (stdout);
    }
  else
    {
      mycutg = NULL;
      printf ("could not read 2dwin file %s\n", winname);
      printf ("TFile error number: %i\n", f->GetErrno ());
      printf ("abort\n\n");
      exit (-1);
    };

  /* done */

  f->Delete ();
  f->Close ();
  return (mycutg);

}


/*----------------------------------------------------*/

TH1D *
mkTH1D (char *str1, char *str2, int nn, double lo, double hi)
{
  TH1D *tmppt;

  tmppt = new TH1D (str1, str2, nn, lo, hi);
  printf ("Created Object \"%s\", %p\n", str1, tmppt);

  return (tmppt);

}

/*----------------------------------------------------*/

TH2F *
mkTH2F (char *str1, char *str2, int nn1, double lo1, double hi1, int nn2, double lo2, double hi2)
{
  TH2F *tmppt;

  tmppt = new TH2F (str1, str2, nn1, lo1, hi1, nn2, lo2, hi2);
  printf ("Created Object \"%s\", %p\n", str1, tmppt);

  return (tmppt);

}

/* ----------------------------------------------------------------- */

int
ctkROOTSetup ()
{

  /* declarations */

  char str1[STRLEN], str2[STRLEN];
  int i;

  printf ("\n");
  printf ("ctkROOTSetup: start\n");
  printf ("\n");

  f1 = new TFile ((char *) Pars.rootFileName, "RECREATE");
  f1->Print ();
  gDirectory->pwd ();

  /* make the 1D root spectra */

  sprintf (str1, "polang");
  sprintf (str2, "polar angles");
  polAngle = mkTH1D (str1, str2, 1800, 0, 180);
  polAngle->SetXTitle (str1);

  sprintf (str1, "EXTtac");
  sprintf (str2, "BGS vs tracked g-rays");
  EXTtac = mkTH1D (str1, str2, 2048, -2048, 2048);
  EXTtac->SetXTitle (str1);

  sprintf (str1, "radius");
  sprintf (str2, "radius");
  radius = mkTH1D (str1, str2, 4096, 0, 30);
  radius->SetXTitle (str1);

  sprintf (str1, "dcx2vsradius");
  sprintf (str2, "dcx2vsradius");
  dcx2vsradius = mkTH1D (str1, str2, 4096, 0, 30);
  dcx2vsradius->SetXTitle (str1);

  sprintf (str1, "dcx2vsradius_norm");
  sprintf (str2, "dcx2vsradius_norm");
  dcx2vsradius_norm = mkTH1D (str1, str2, 4096, 0, 30);
  dcx2vsradius_norm->SetXTitle (str1);

  sprintf (str1, "dopplerfac");
  sprintf (str2, "doppler factors");
  dopfac = mkTH1D (str1, str2, 1000, 0.5, 1.5);
  dopfac->SetXTitle (str1);

  sprintf (str1, "fm");
  sprintf (str2, "figure of merit");
  fmsp = mkTH1D (str1, str2, 1024, 0, 2.1);
  fmsp->SetXTitle (str1);

  sprintf (str1, "dcx2fm");
  sprintf (str2, "dcx2 figure of merit");
  dcx2 = mkTH1D (str1, str2, 1024, 0, 20.0);
  dcx2->SetXTitle (str1);

  sprintf (str1, "sumTrackE");
  sprintf (str2, "tracked energy sum (subject to gates in chatfile)");
  sumTrackE = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  sumTrackE->SetXTitle (str1);

  sprintf (str1, "BGSEsi");
  sprintf (str2, "BGS Si energies");
  BGSEsi = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  BGSEsi->SetXTitle (str1);

  sprintf (str1, "EXT_tac");
  sprintf (str2, "EXT_tac");
  sEXT_tac = mkTH1D (str1, str2, LONGLEN / 2, -LONGLEN / 2, LONGLEN / 2);
  sprintf (str1, "time");
  sEXT_tac->SetXTitle (str1);

  sprintf (str1, "G4_obs");
  sprintf (str2, "G4 Observed energy");
  G4_obs = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  G4_obs->SetXTitle (str1);

  sprintf (str1, "G4_emit");
  sprintf (str2, "G4 emitted energy");
  G4_emit = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  G4_emit->SetXTitle (str1);

  sprintf (str1, "sumehi");
  sprintf (str2, "per detector energy sum");
  sumehi = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  sumehi->SetXTitle (str1);

  sprintf (str1, "decompSumE");
  sprintf (str2, "orig energy sum");
  decompSumE = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  decompSumE->SetXTitle (str1);

  for (i = 0; i < MAXDETNO; i++)
    {
      sprintf (str1, "ehi%3.3i", i);
      sprintf (str2, "ehi%3.3i @ %6.2f deg", i, Pars.detpolang[i]);
      ehi[i] = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
      sprintf (str1, "(keV)");
      ehi[i]->SetXTitle (str1);

      sprintf (str1, "radius_c_%3.3i", i);
      sprintf (str2, "radius_c_%3.3i", i);
      radius_c[i] = mkTH1D (str1, str2, 4096, 0, 30);
      radius_c[i]->SetXTitle (str1);


    };


  sprintf (str1, "FPGAE");
  sprintf (str2, "FPGAE");
  FPGAE = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN);
  sprintf (str1, "(keV)");
  FPGAE->SetXTitle (str1);
  egamBinWidth = (float) FPGAE->GetBinWidth (1);

  sprintf (str1, "hitpat");
  sprintf (str2, "hitpat");
  hitpat = mkTH1D (str1, str2, 200, 1, 200);
  sprintf (str1, "det number");
  hitpat->SetXTitle (str1);


  /* make the 2D root spectra */

  sprintf (str1, "fomXe");
  sprintf (str2, "fomXe");
  fomXe = mkTH2F (str1, str2, GGMAX, 1, GGMAX, 200, 0, 2.0);
  sprintf (str1, "e");
  fomXe->SetXTitle (str1);
  sprintf (str1, "fom");
  fomXe->SetYTitle (str1);

  sprintf (str1, "S800_PID");
  sprintf (str2, "S800_PID");
  S800_PID =
    mkTH2F (str1, str2, Pars.S800_PID_NX, Pars.S800_PID_XLO, Pars.S800_PID_XHI, Pars.S800_PID_NY, Pars.S800_PID_YLO,
            Pars.S800_PID_YHI);
  sprintf (str1, "x");
  S800_PID->SetXTitle (str1);
  sprintf (str1, "y");
  S800_PID->SetYTitle (str1);


  sprintf (str1, "rad_e");
  sprintf (str2, "rad_e");
  rad_e = mkTH2F (str1, str2, 2048, 0, 30, LONGLEN / 4, 1, LONGLEN / 2);
  sprintf (str1, "radius");
  rad_e->SetXTitle (str1);
  sprintf (str1, "E (keV)");
  rad_e->SetYTitle (str1);


  /* star map of GRETA */

  sprintf (str1, "SMAP_allhits");
  sprintf (str2, "SMAP_allhits");
  SMAP_allhits = mkTH2F (str1, str2, Pars.nBinCh2d, Pars.minazi, Pars.maxazi, Pars.nBinCh2d, Pars.minpol, Pars.maxpol);
  sprintf (str1, "horizontal");
  SMAP_allhits->SetXTitle (str1);
  sprintf (str1, "vertical");
  SMAP_allhits->SetYTitle (str1);

  sprintf (str1, "SMAP_firsthits");
  sprintf (str2, "SMAP_firsthits");
  SMAP_firsthits =
    mkTH2F (str1, str2, Pars.nBinCh2d, Pars.minazi, Pars.maxazi, Pars.nBinCh2d, Pars.minpol, Pars.maxpol);
  sprintf (str1, "horizontal");
  SMAP_firsthits->SetXTitle (str1);
  sprintf (str1, "vertical");
  SMAP_firsthits->SetYTitle (str1);

  sprintf (str1, "gg");
  sprintf (str2, "gg");
  gg = mkTH2F (str1, str2, GGMAX, 1, GGMAX, GGMAX, 1, GGMAX);
  sprintf (str1, "g1");
  gg->SetXTitle (str1);
  sprintf (str1, "g2");
  gg->SetYTitle (str1);

  sprintf (str1, "gg_fpga");
  sprintf (str2, "gg_fpga");
  gg_fpga = mkTH2F (str1, str2, GGMAX, 1, GGMAX, GGMAX, 1, GGMAX);
  sprintf (str1, "g1");
  gg_fpga->SetXTitle (str1);
  sprintf (str1, "g2");
  gg_fpga->SetYTitle (str1);

  sprintf (str1, "gg_decomp");
  sprintf (str2, "gg_decomp");
  gg_decomp = mkTH2F (str1, str2, GGMAX, 1, GGMAX, GGMAX, 1, GGMAX);
  sprintf (str1, "g1");
  gg_decomp->SetXTitle (str1);
  sprintf (str1, "g2");
  gg_decomp->SetYTitle (str1);

  /* read 2D PID window */

  if (Pars.havePIDwin)
    PIDwin = rd2dwin (Pars.PIDwinfile);

  /* list what we have */

  wlist = gDirectory->GetList ();
  wlist->Print ();

  /* some counting */

  nPID = 0;
  nPIDwin = 0;
  nGam = 0;

  /* done */

  printf ("\n");
  printf ("ctkROOTSetup: done\n");
  printf ("\n");
  return (0);

}

/* ----------------------------------------------------------------- */

int
ctkROOTexit ()
{

  /* declarations */

  int dim, i;
  float rr[LONGLEN];
  double d1;
  char str[512];

  /* prototypes */

  int ctkWsig (TH1D *, char *, float);
  int wr_spe (char *, int *, float *);

  /*-----------*/
  /* normalize */
  /*-----------*/

  dcx2vsradius->Divide (dcx2vsradius_norm);


  /*---------------------*/
  /* write root file out */
  /*---------------------*/

  printf ("\n");
  printf ("ctkROOTexit: start\n");
  printf ("\n");
  gDirectory->pwd ();

  f1->Write ();

  wlist = gDirectory->GetList ();
  wlist->Print ();

  printf ("writing all spectra to \"%s\"...", Pars.rootFileName);
  wlist->Write (0, TObject::kOverwrite);
  printf ("DONE!\n");
  printf ("\n");
  fflush (stdout);

  /*-----------------------------*/
  /* do the statistical analysis */
  /*-----------------------------*/

  /* write the tracked spectrum in spe format as well */
  /* sumTrackE = mkTH1D (str1, str2, LONGLEN, 1, LONGLEN); */

  dim = LONGLEN;
  for (i = 0; i < dim; i++)
    {
      rr[i] = (float) sumTrackE->GetBinContent (i);;
    };
  sprintf (str, "sumTrackE.spe");
  wr_spe (str, &dim, rr);

  dim = LONGLEN;
  for (i = 0; i < dim; i++)
    {
      rr[i] = (float) sumehi->GetBinContent (i);;
    };
  sprintf (str, "sumehi.spe");
  wr_spe (str, &dim, rr);

  /* S800 stats */

  d1 = 100 * (float) nPIDwin / (float) nPID;
  printf ("nPID=%li, nPIDwin=%li or %9.2f%%\n", nPID, nPIDwin, (float) d1);
  d1 = 100 * (float) nGam / (float) nPIDwin;
  printf ("nGam=%li or %9.2f%% of nPIDwin\n", nGam, (float) d1);

  /* done */

  printf ("\n");
  printf ("ctkROOTexit: done []\n");
  printf ("\n");
  return (0);

}

/* ----------------------------------------------------------------- */

int
ctkROOTBin (TRACK_STRUCT * track, float polang[MAXCLUSTERHITS], float doppler_factor[MAXCLUSTERHITS],
            BGS_STRUCT * bgs, long long int EXT_tac)
{

  /* declarations */

  double d1, d2, val;
  PAYLOAD *ptinp;
  GEBDATA *ptgd;
  CRYS_INTPTS *gtinp;
  int i, k, j, l, crystalno, moduleno, detno, PIDok, nS800, i1;
  float sX, sY, polAng, aziAng;
  float x1, y1, z1;
  S800_PHYSICSDATA *ptDirk;
  float rr, dp, rn;
  COOR coor;
  double CosDop;
  double xsin;
  double ysin;
  double s800_theta;
  double s800_phi;
  double dbeta;
  double Gamma;
  double dp_p;
  double TheGamma;
  static int nn = 0, ngrh = 0;
  TRACKED_GAMMA_HIT *grh;
  int ngam;
  double gam[100];
  char str[512];
  FILE *fp;
  float g4eg[MAXL], g4xx[MAXL], g4yy[MAXL], g4zz[MAXL];
  int g4nhits, g4id[MAXL];
  float g4egamma, g4doppler_factor;
  int g4evno;

//  printf("track->n=%i\n",track->n);
//  printf("\n",);
//  printf("\n",);
//  fflush(stdout);  

  if (Pars.nprint > 0)
    {

      /* open in append mode so we can add */
      /* doppler correction information */

      sprintf (str, "trackedEvent_%3.3i.txt", Pars.evno);
      fp = fopen (str, "a");
      fprintf (fp, "entered ctkROOTBin.c\n");
    }

  /* if we have Geant4 simulated data */
  /* process it */

  ptinp = track->payload;
  ptgd = track->gd;
  for (k = 0; k < track->n; k++)
    {

      if (ptgd->type == GEB_TYPE_G4SIM)
        {

#if(0)
          printf ("found Geant4 simulated data\n");
#endif

          i1 = 0;
          memcpy (&g4nhits, (char *) ptinp + i1, sizeof (int));
          i1 += sizeof (int);
          memcpy (&g4egamma, (char *) ptinp + i1, sizeof (float));
          i1 += sizeof (float);
          memcpy (&g4evno, (char *) ptinp + i1, sizeof (int));
          i1 += sizeof (int);
          memcpy (&g4doppler_factor, (char *) ptinp + i1, sizeof (float));
          i1 += sizeof (float);
#if(0)
          printf ("nhits=%3i, e=%9.2f, evno= %i, doppler_factor=%f\n", g4nhits, g4egamma, g4evno, g4doppler_factor);
#endif

          val = 0;
          for (i = 0; i < g4nhits; i++)
            {
              memcpy (&g4id[i], (char *) ptinp + i1, sizeof (int));
              i1 += sizeof (int);
              memcpy (&g4eg[i], (char *) ptinp + i1, sizeof (float));
              i1 += sizeof (float);
              val += g4eg[i];
              memcpy (&g4xx[i], (char *) ptinp + i1, sizeof (float));
              i1 += sizeof (float);
              memcpy (&g4yy[i], (char *) ptinp + i1, sizeof (float));
              i1 += sizeof (float);
              memcpy (&g4zz[i], (char *) ptinp + i1, sizeof (float));
              i1 += sizeof (float);
            };

#if(0)
          for (i = 0; i < g4nhits; i++)
            {
              printf ("%i ", g4id[i]);
              printf ("%f ", g4eg[i]);
              printf ("%f ", g4xx[i]);
              printf ("%f ", g4yy[i]);
              printf ("%f\n", g4zz[i]);
            };
#endif

          /* now bin */

          val /= g4doppler_factor;
          if (val > 0 && val < LONGLEN)
            G4_obs->Fill ((double) val, 1);

          val = g4egamma;
          val /= g4doppler_factor;
          if (val > 0 && val < LONGLEN)
            G4_emit->Fill ((double) val, 1);


//          if (1)            exit (0);

        };

      ptinp++;
      ptgd++;
    };



  /* plug out at the tracked data */

  ngrh = 0;
  ptinp = track->payload;
  ptgd = track->gd;
  for (k = 0; k < track->n; k++)
    {

      if (ptgd->type == GEB_TYPE_TRACK)
        {

          /* cast data into tracked data structure */

          ngrh++;
          grh = (TRACKED_GAMMA_HIT *) ptinp;

        };

      ptinp++;
      ptgd++;
    };

  if (ngrh != 1)
    return (0);

  /* limit fom range for technical reasons */

  for (i = 0; i < grh->ngam; i++)
    {
//      printf("grh->gr[%i].fom=%f\n", i,grh->gr[i].fom);
//      fflush(stdout);
      if (grh->gr[i].fom >= 1.99)
        grh->gr[i].fom = 1.99;
    };

#if(0)

  /* code to see what we were passed */

  ptinp = track->payload;
  ptgd = track->gd;
  printf ("ctkROOTBin was passed (track->n=%i):\n", track->n);
  for (k = 0; k < track->n; k++)
    {
      printf ("1: k=%i, ptgd->type =%i, track->n=%i\n", k, ptgd->type, track->n);
      fflush (stdout);

      /* next */

      ptinp++;
      ptgd++;

    };

  nn++;
  if (nn >= 2)
    exit (0);

#endif

/* require external TS to be  */
/* in range or return */

  if (Pars.gateexternal)
    {
      if (EXT_tac < Pars.EXT_tac_lo)
        {

          if (Pars.nprint > 0)
            {
              fprintf (fp, "failed EXT_tac low limit\n");
              fclose (fp);
            };
          return (0);
        }

      if (EXT_tac > Pars.EXT_tac_hi)
        {
          if (Pars.nprint > 0)
            {
              fprintf (fp, "failed EXT_tac hi limit\n");
              fclose (fp);
            };
          return (0);
        }
    };

  if (Pars.nprint > 0)
    fprintf (fp, "tac values OK\n");

/* bin S800 IC 2D (PID); count as well */

  nS800 = 0;
  ptinp = track->payload;
  ptgd = track->gd;
  PIDok = 0;
  for (k = 0; k < track->n; k++)
    {

      if (ptgd->type == GEB_TYPE_S800PHYSDATA)
        {
          nS800++;
          nPID++;

          /* cast data into Dirks structure */

          ptDirk = (S800_PHYSICSDATA *) ptinp;

          /* update 2D particle ID */

//          d1 = ptDirk->tof_rfe1;
//          d1 = ptDirk->tof_obje1;
//          d2 = ptDirk->ic_de;
//          d2 = ptDirk->ic_sum;

          d1 = ptDirk->tof_obje1;
          d2 = ptDirk->ic_de;


          /* check if inside window */

          if (Pars.havePIDwin)
            {
              if (PIDwin->IsInside (d1, d2))
                {
                  PIDok++;
                  if (d1 > Pars.S800_PID_XLO && d1 < Pars.S800_PID_XHI)
                    if (d2 > Pars.S800_PID_YLO && d2 < Pars.S800_PID_YHI)
                      S800_PID->Fill (d1, d2);
//             printf("inside %f %f; ", (float)d1, (float)d2);
//             printf("x: %f to %f, ",Pars.S800_PID_XLO,Pars.S800_PID_XHI);
//             printf("y: %f to %f\n",Pars.S800_PID_YLO,Pars.S800_PID_YHI);
                }
              else
                {
//              printf("outside %f %f; ", (float)d1, (float)d2);
//             printf("x: %f to %f, ",Pars.S800_PID_XLO,Pars.S800_PID_XHI);
//             printf("y: %f to %f\n",Pars.S800_PID_YLO,Pars.S800_PID_YHI);
                };
            }
          else
            {

//             printf("everything %f %f; ", (float)d1, (float)d2);
//             printf("x: %f to %f, ",Pars.S800_PID_XLO,Pars.S800_PID_XHI);
//             printf("y: %f to %f\n",Pars.S800_PID_YLO,Pars.S800_PID_YHI);

              /* bin everything */

              if (d1 > Pars.S800_PID_XLO && d1 < Pars.S800_PID_XHI)
                if (d2 > Pars.S800_PID_YLO && d2 < Pars.S800_PID_YHI)
                  S800_PID->Fill (d1, d2);

            };

        };

      ptinp++;
      ptgd++;
    };

  /* if PID window is there, check it */

  if (Pars.havePIDwin)
    {
      if (PIDok != 1)
        {
          if (Pars.nprint > 0)
            {
              fprintf (fp, "failed PIDok\n");
              fclose (fp);
            };
          return (0);
        }
    };

  nPIDwin++;
  if (Pars.nprint > 0)
    fprintf (fp, "PID window OK\n");


//printf("ngrh=%i\n");
//print_tracked_gamma_rays (stdout, grh);
//if(1)exit(0);

  assert (grh->ngam < MAX_GAMMA_RAYS);

  /*--------------------------------------------------*/
  /* do the doppler corrections of all gamma energies */
  /*--------------------------------------------------*/


  if (nS800 == 0 || Pars.simpleDopplerCor)
    {

      if (Pars.nprint > 0)
        fprintf (fp, "simple non s800 doppler corection\n");

      /* simple non-S800 case */

      for (i = 0; i < grh->ngam; i++)
        if (grh->gr[i].tracked)
          {

            /* find poler angles relative to beam direction */

            rr = grh->gr[i].x0 * grh->gr[i].x0 + grh->gr[i].y0 * grh->gr[i].y0 + grh->gr[i].z0 * grh->gr[i].z0;
            rr = sqrtf (rr);

            dp = (grh->gr[i].x0 * Pars.beamdir[0] +
                  grh->gr[i].y0 * Pars.beamdir[1] + grh->gr[i].z0 * Pars.beamdir[2]) / rr;

            if (dp < -1.0)
              dp = -1.0;
            if (dp > 1.0)
              dp = 1.0;
            polang[i] = acosf (dp);


            /* find (simple) doppler correction factor */
            /* for this gamma ray 'i' of 'nClusters' */

            rr = 1.0 - Pars.beta * Pars.beta;
            doppler_factor[i] = sqrt (rr) / (1.0 - Pars.beta * cos (polang[i]));

            if (Pars.nprint > 0)
              {
                fprintf (fp, "gamma ray # %i\n", i);
                fprintf (fp, "x=%6.2f ", grh->gr[i].x0);
                fprintf (fp, "y=%6.2f ", grh->gr[i].y0);
                fprintf (fp, "z=%6.2f \n", grh->gr[i].z0);
                fprintf (fp, "detection vector: (%9.6f ", grh->gr[i].x0 / rr);
                fprintf (fp, "%9.6f ", grh->gr[i].y0 / rr);
                fprintf (fp, "%9.6f) ", grh->gr[i].z0 / rr);
                x1 = 0;
                y1 = 0;
                z1 = 1;
                fprintf (fp, "and s800 (%9.6f ", x1);
                fprintf (fp, "%9.6f ", y1);
                fprintf (fp, "%9.6f)\n", z1);

                fprintf (fp, "dp= %10.7f\n", dp);

                fprintf (fp, "polar angle is %10.5f deg, %10.5f  rad, dp=%f\n", polang[i] * 180. / M_PI, polang[i], dp);
                fprintf (fp, "standard: doppler correction for gamm # %i = %f\n", i, doppler_factor[i]);

              }


          };

    }
  else

    {

      /* have S800 data, so we need to do proper */
      /* doppler correction using S800 data */
      /* The code for this calculation was */
      /* proudly stolen from Dirk */

      if (Pars.nprint > 0)
        fprintf (fp, "full s800 doppler corection\n");

      if (Pars.nprint > 0)
        print_tracked_gamma_rays (fp, grh);

      for (i = 0; i < grh->ngam; i++)
        if (grh->gr[i].tracked)
          {

            /* then with corrections we have */
            /* the first interaction point (we use cm) */
            /* note: we have already done the rotation */
            /* and translation, assuming target is at center */
            /* in the tracking task, so we just have to */
            /* apply the corrections here */

// this gives worse resolution even though that is how Dirk
// suggested using the corrections; don't know why
            coor.x = grh->gr[i].x0 + Pars.target_x;
// think this is the way it should be, but
//            coor.y = grh->gr[i].y0 + Pars.target_y - ptDirk->yta * 100.0;
// to get the same numbers as Dirk, i need to use this sign:
            coor.y = grh->gr[i].y0 + Pars.target_y + Pars.ytascale * ptDirk->yta * 100.0;
            coor.z = grh->gr[i].z0 + Pars.target_z;


            if (Pars.nprint > 0)
              {
                fprintf (fp, "Pars.target_x=%6.2f\n", (float) Pars.target_x);
                fprintf (fp, "Pars.target_y=%6.2f\n", (float) Pars.target_y);
                fprintf (fp, "Pars.target_z=%6.2f\n", (float) Pars.target_z);
                fprintf (fp, "Pars.ytascale=%6.2f\n", (float) Pars.ytascale);
                fprintf (fp, "Pars.ytascale*ptDirk->yta*100.0=%6.2f\n", Pars.ytascale * ptDirk->yta * 100.0);
                fprintf (fp, "target pos after target offset and S800 correction\n");
                fprintf (fp, "x=%6.2f ", (float) coor.x);
                fprintf (fp, "y=%6.2f ", (float) coor.y);
                fprintf (fp, "z=%6.2f ", (float) coor.z);
                fprintf (fp, "\n");
              }

            xsin = Pars.greta_ata / 180. * M_PI;
            ysin = Pars.greta_bta / 180. * M_PI;
            if (Pars.nprint > 0)
              {
                fprintf (fp, "a: ");
                fprintf (fp, "xsin=%f; ", (float) xsin);
                fprintf (fp, "ysin=%f", (float) ysin);
                fprintf (fp, "\n with: ");
                fprintf (fp, "Pars.greta_ata=%f ", Pars.greta_ata);
                fprintf (fp, "Pars.greta_bta=%f", Pars.greta_bta);
                fprintf (fp, "\n");
              }

            xsin += ptDirk->ata;
            ysin += ptDirk->bta;
            if (Pars.nprint > 0)
              {
                fprintf (fp, "b: ");
                fprintf (fp, "xsin=%f; ", (float) xsin);
                fprintf (fp, "ysin=%f", (float) ysin);
                fprintf (fp, "\n");
              }

            xsin = sin (xsin);
            ysin = sin (ysin);
            if (Pars.nprint > 0)
              {
                fprintf (fp, "c: ");
                fprintf (fp, "xsin=%f; ", (float) xsin);
                fprintf (fp, "ysin=%f", (float) ysin);
                fprintf (fp, "\n");
              }

            if (xsin > (double) 0 && ysin > (double) 0)
              s800_phi = -atan (ysin / xsin);
            else if (xsin < (double) 0 && ysin > (double) 0)
              s800_phi = -M_PI + atan (ysin / fabs (xsin));
            else if (xsin < (double) 0 && ysin < (double) 0)
              s800_phi = -M_PI - atan (fabs (ysin) / fabs (xsin));
            else if (xsin > (double) 0 && ysin < (double) 0)
              s800_phi = -2 * M_PI + atan (fabs (ysin) / xsin);
            else
              s800_phi = 0.0;
            s800_theta = asin (sqrt (xsin * xsin + ysin * ysin));
            if (Pars.nprint > 0)
              {
                fprintf (fp, "s800_theta = %f rad or %f deg\n", s800_theta, s800_theta * 180 / M_PI);
                fprintf (fp, "s800_phi = %f rad or %f deg \n", s800_phi, s800_phi * 180 / M_PI);
              }

            /* calculate the polar and azimuth */
            /* angle for first interaction point */

            rr = sqrtf (coor.x * coor.x + coor.y * coor.y + coor.z * coor.z);
            if (Pars.nprint > 0)
              {
                fprintf (fp, "detection vector: (%9.6f ", coor.x / rr);
                fprintf (fp, "%9.6f ", coor.y / rr);
                fprintf (fp, "%9.6f) ", coor.z / rr);
                x1 = sin (s800_theta) * cos (s800_phi);
                y1 = sin (s800_theta) * sin (s800_phi);;
                z1 = cos (s800_theta);
                fprintf (fp, "and s800 (%9.6f ", x1);
                fprintf (fp, "%9.6f ", y1);
                fprintf (fp, "%9.6f)\n", z1);
              }
            coor.theta = acosf (coor.z / rr);

            if (coor.x > 0 && coor.y > 0)
              coor.phi = atanf (coor.y / coor.x);
            else if (coor.x < 0 && coor.y > 0)
              coor.phi = M_PI - atanf (coor.y / fabs (coor.x));
            else if (coor.x < 0 && coor.y < 0)
              coor.phi = M_PI + atanf (coor.y / coor.x);
            else if (coor.x > 0 && coor.y < 0)
              coor.phi = 2 * M_PI - atanf (fabs (coor.y) / coor.x);
            else
              coor.phi = 0;

            if (Pars.nprint > 0)
              {
                fprintf (fp, "coor.theta=%f\n", coor.theta);
                fprintf (fp, "coor.phi=%f\n", coor.phi);
              }


            CosDop =
              sin (coor.theta) * sin (s800_theta) *
              (sin (coor.phi) * sin (s800_phi) + cos (coor.phi) * cos (s800_phi)) + cos (coor.theta) * cos (s800_theta);
            if (Pars.nprint > 0)
              {
                fprintf (fp, "CosDop=%f\n", CosDop);
              }
            dbeta = Pars.beta;
            if (Pars.nprint > 0)
              {
                printf ("fp,a:dbeta=%f\n", dbeta);
              }
            Gamma = 1.0 / sqrt (1.0 - dbeta * dbeta);
            dp_p = Gamma / (1 + Gamma) * ptDirk->dta;
            dbeta *= (1.0 + dp_p / (Gamma * Gamma));
            if (Pars.nprint > 0)
              {
                fprintf (fp, "Gamma=%f\n", Gamma);
                fprintf (fp, "dp_p=%f\n", dp_p);
                fprintf (fp, "b:dbeta=%f\n", dbeta);
              }


            TheGamma = 1.0 / sqrt (1.0 - dbeta * dbeta);
            if (Pars.nprint > 0)
              {
                fprintf (fp, "TheGamma=%f\n", TheGamma);
              }

            /* goal:: is to find the doppler_factor[i] */

            doppler_factor[i] = 1.0 / (float) (TheGamma * (1 - dbeta * CosDop));
            polang[i] = coor.theta;
            if (Pars.nprint > 0)
              {
                fprintf (fp, "S800: doppler correction for %i= %f\n", i, doppler_factor[i]);
              }

          };


    };
  nn++;
#if(DDEBUG)
  if (nn >= 1)
    exit (0);
#endif

  /* now apply the doppler correction */
  /* factors calculated above */

  if (Pars.nprint > 0)
    fprintf (fp, "we have %i tracked gamma rays\n", grh->ngam);

  for (i = 0; i < grh->ngam; i++)
    if (grh->gr[i].tracked)
      {
        if (Pars.nprint > 0)
          {
            fprintf (fp, "gamma ray # %i\n", i);
            fprintf (fp, "LAB: esum=%9.4f ", grh->gr[i].esum);
            fprintf (fp, "e0=%12.7f ", grh->gr[i].e0);
            fprintf (fp, "e1=%12.7f\n", grh->gr[i].e1);
            fprintf (fp, "doppler_factor[%i]= %9.5f or 1/%9.5f, corrected:\n", i, doppler_factor[i],
                     1.0 / doppler_factor[i]);
          }

        grh->gr[i].esum /= doppler_factor[i];
        grh->gr[i].e0 /= doppler_factor[i];
        grh->gr[i].e1 /= doppler_factor[i];

        if (Pars.nprint > 0)
          {
            fprintf (fp, "CMS: esum=%9.4f ", grh->gr[i].esum);
            fprintf (fp, "e0=%12.7f ", grh->gr[i].e0);
            fprintf (fp, "e1=%12.7f\n", grh->gr[i].e1);
          }

      };


  if (EXT_tac > -LONGLEN / 2)
    if (EXT_tac < LONGLEN / 2)
      sEXT_tac->Fill ((double) EXT_tac, 1);


  /*----------------------------------------------*/
  /* fill the SMAP_allhits GRETINA hitpattern map */
  /*----------------------------------------------*/

  ptinp = track->payload;
  ptgd = track->gd;
  for (k = 0; k < track->n; k++)
    {
//    printf ("2:k=%i, ptgd->type =%i\n", k, ptgd->type);
//    fflush (stdout);

      if (ptgd->type == GEB_TYPE_DECOMP)
        {
          gtinp = (CRYS_INTPTS *) ptinp;
//        printf("hi1\n");
          if (Pars.enabled[Pars.detNo[k]])
            {
//            printf("hi2\n");

              crystalno = (gtinp->crystal_id & 0x0003);
              moduleno = ((gtinp->crystal_id & 0xfffc) >> 2);
              detno = moduleno * 4 + crystalno;

              for (j = 0; j < gtinp->num; j++)
                {


                  polAng = findPolarFromCartesian (gtinp->intpts[j].x, gtinp->intpts[j].y, gtinp->intpts[j].z, &rr);
                  aziAng = findAzimuthFromCartesian (gtinp->intpts[j].x, gtinp->intpts[j].y, gtinp->intpts[j].z);
                  radius->Fill ((double) rr, 1);

                  radius_c[detno]->Fill ((double) rr, 1);

                  if (Pars.nprint > 0)
                    {
                      printf ("%2i: ", j);
                      printf ("x=%f ", gtinp->intpts[j].x);
                      printf ("y=%f ", gtinp->intpts[j].y);
                      printf ("z=%f; ", gtinp->intpts[j].z);
                      printf ("radius=%f; ", (float) rr);
                      printf ("aziAng=%f ", aziAng / rad2deg);
                      printf ("polAng=%f\n", polAng / rad2deg);
                    }




                  mean_pol += (double) polAng / rad2deg;
                  mean_azi += polAng / rad2deg;
                  mean_n++;

                  /* SMAP coordinates */

                  sX = aziAng * sinf (polAng) / rad2deg + 180;
                  sY = polAng / rad2deg;        /* + 1.5; */

#if(1)
                  if (Pars.nprint > 0)
                    {
                      printf ("sX,sY= ( %11.6f , %11.6f )\n", sX, sY);
                      fflush (stdout);
                    }
#endif

                  /* update */

                  if (rr < Pars.rmax && rr > Pars.rmin)
                    {
                      if (sX >= 0 && sX <= 360 && sY >= 0 && sY <= 180)
                        SMAP_allhits->Fill (sX, sY, 1);
                      else
                        {
                          printf ("error: sX,sY= ( %11.6f , %11.6f )\n", sX, sY);
//                          exit (1);
                        };
                    }

                };



            };

        };

      /* next */

      ptinp++;
      ptgd++;

    };


  /*------------------------------------*/
  /* bin original summed energy spectrum */
  /*------------------------------------*/

  ptinp = track->payload;
  ptgd = track->gd;
  ngam = 0;
  for (k = 0; k < track->n; k++)
    {

      if (ptgd->type == GEB_TYPE_DECOMP)
        {
          gtinp = (CRYS_INTPTS *) ptinp;
          if (Pars.enabled[Pars.detNo[k]])
            {
              d2 = 0;
              for (j = 0; j < gtinp->num; j++)
                {
                  d2 += gtinp->intpts[j].e;
                };
              d1 = (float) rand () / RAND_MAX;
              d1 = (1.0 * d2) + egamBinWidth * d1;
              decompSumE->Fill (d1, 1);

              gam[ngam] = d1;
              ngam++;


              crystalno = (gtinp->crystal_id & 0x0003);
              moduleno = ((gtinp->crystal_id & 0xfffc) >> 2);
              detno = moduleno * 4 + crystalno;
              d1 = gtinp->tot_e / Pars.detdopfac[detno];
              ehi[detno]->Fill (d1, 1);
              sumehi->Fill (d1, 1);

              hitpat->Fill ((double) detno, 1);

              /*printf("%i\n", i1); */
            };

        };

      /* next */

      ptinp++;
      ptgd++;
    };

  /* fill gg_decomp matrix */

  if (ngam >= 2)
    for (k = 0; k < ngam; k++)
      for (l = 0; l < ngam; l++)
        if (k != l)
          gg_decomp->Fill (gam[k], gam[l], 1);


  /* decomposition Chi2 vs radius */

  ptinp = track->payload;
  ptgd = track->gd;
  for (k = 0; k < track->n; k++)
    {
      if (ptgd->type == GEB_TYPE_DECOMP)
        {
          gtinp = (CRYS_INTPTS *) ptinp;
          if (Pars.enabled[Pars.detNo[k]])
            {
              for (j = 0; j < gtinp->num; j++)
                {
                  d2 = gtinp->intpts[j].x * gtinp->intpts[j].x
                    + gtinp->intpts[j].y * gtinp->intpts[j].y + gtinp->intpts[j].z * gtinp->intpts[j].z;
                  d2 = sqrt (d2);
//          printf("%f,%f\n", d2,gtinp->norm_chisq);
                  dcx2vsradius->Fill (d2, gtinp->norm_chisq);
                  dcx2vsradius_norm->Fill (d2, 1);
                };
            };
        };
      ptinp++;
      ptgd++;
    };


  /*------------------------------------*/
  /* bin FPGA energy etc spectra */
  /*------------------------------------*/

  ptinp = track->payload;
  ptgd = track->gd;
  ngam = 0;
  for (k = 0; k < track->n; k++)
    {

      if (ptgd->type == GEB_TYPE_DECOMP)
        {
          gtinp = (CRYS_INTPTS *) ptinp;
          if (Pars.enabled[Pars.detNo[k]])
            {
              /* fpga energy */

              /*printf ("ptinp-tot_e>=%f\n", gtinp->tot_e); */

              d1 = 1.0 * gtinp->tot_e;
              FPGAE->Fill (d1, 1);

              /* decomposition FOM */

              dcx2->Fill (gtinp->norm_chisq, 1);

              gam[ngam] = gtinp->tot_e;
              ngam++;

            };
        };

      /* next */

      ptinp++;
      ptgd++;
    };

  /* fill gg_fpga matrix */

  if (ngam >= 2)
    for (k = 0; k < ngam; k++)
      for (l = 0; l < ngam; l++)
        if (k != l)
          gg_fpga->Fill (gam[k], gam[l], 1);


  /*--------------------------*/
  /* fill tracked data arrays */
  /*--------------------------*/

  for (i = 0; i < grh->ngam; i++)
    if (grh->gr[i].tracked)
      if (grh->gr[i].ndet >= Pars.ndetlimlo && grh->gr[i].ndet <= Pars.ndetlimhi)
        {


          d1 = (double) (1000.0 * grh->gr[i].esum);
          if (grh->gr[i].fom < 2.0 && grh->gr[i].fom >= 0)
            if (d1 > 0 && d1 < GGMAX)
              fomXe->Fill (d1, grh->gr[i].fom, 1);

        };

  for (i = 0; i < grh->ngam; i++)
    if (grh->gr[i].tracked)
      if (grh->gr[i].ndet >= Pars.ndetlimlo && grh->gr[i].ndet <= Pars.ndetlimhi)

        {
//        printf("grh->gr[i].ndet=%i\n",grh->gr[i].ndet);
//        fflush(stdout);
          if (grh->gr[i].fom >= Pars.fomlo[grh->gr[i].ndet] && grh->gr[i].fom <= Pars.fomhi[grh->gr[i].ndet])
            if (grh->gr[i].ndet >= Pars.ndetlimlo && grh->gr[i].ndet <= Pars.ndetlimhi)
              {

                if (grh->gr[i].fom >= 0 && grh->gr[i].fom < 2.0)
                  fmsp->Fill (grh->gr[i].fom, 1);

                d1 = (double) (1000.0 * grh->gr[i].esum);
                if (d1 < 0)
                  d1 = 0;
                if (d1 > LONGLEN)
                  d1 = LONGLEN-1;

                sumTrackE->Fill (d1, 1);
                rad_e->Fill (rr, d1, 1.0);
                nGam++;

                polAngle->Fill (polang[i] / RAD2DEG, 1);
                dopfac->Fill (doppler_factor[i], 1);

              };
        };

  /* gamma gamma matrix */

  if (grh->ngam >= 2)
    for (i = 0; i < grh->ngam; i++)
      if (grh->gr[i].tracked)
        if (grh->gr[i].ndet >= Pars.ndetlimlo && grh->gr[i].ndet <= Pars.ndetlimhi)
          if (grh->gr[i].fom >= Pars.fomlo[grh->gr[i].ndet] && grh->gr[i].fom <= Pars.fomhi[grh->gr[i].ndet])
            if (grh->gr[i].ndet >= Pars.ndetlimlo && grh->gr[i].ndet <= Pars.ndetlimhi)
              for (j = 0; j < grh->ngam; j++)
                if (i != j)
                  if (grh->gr[j].tracked)
                    if (grh->gr[j].ndet >= Pars.ndetlimlo && grh->gr[j].ndet <= Pars.ndetlimhi)
                      if (grh->gr[j].fom >= Pars.fomlo[grh->gr[j].ndet]
                          && grh->gr[j].fom <= Pars.fomhi[grh->gr[j].ndet])
                        if (grh->gr[j].ndet >= Pars.ndetlimlo && grh->gr[j].ndet <= Pars.ndetlimhi)
                          {

                            rn = (float) rand () / RAND_MAX;
                            rn -= 0.5;
                            d1 = (double) (1000.0 * grh->gr[i].esum + rn);
                            if (d1 < 0)
                              d1 = 0;
                            if (d1 > GGMAX)
                              d1 = GGMAX;

                            rn = (float) rand () / RAND_MAX;
                            rn -= 0.5;
                            d2 = (double) (1000.0 * grh->gr[j].esum + rn);
                            if (d2 < 0)
                              d2 = 0;
                            if (d2 > GGMAX)
                              d2 = GGMAX;


                            gg->Fill (d1, d2, 1.0);
                            gg->Fill (d2, d1, 1.0);

                          };



  /*--------------------------------------------------*/
  /* fill the first interaction point SMAP hitpattern */
  /*--------------------------------------------------*/

  for (i = 0; i < grh->ngam; i++)
    if (grh->gr[i].tracked)
      {

        /* find polar and azimuth angles */


        polAng = findPolarFromCartesian (grh->gr[i].x0, grh->gr[i].y0, grh->gr[i].z0, &rr);
        aziAng = findAzimuthFromCartesian (grh->gr[i].x0, grh->gr[i].y0, grh->gr[i].z0);

//          printf("aziAng=%6.2f\n",aziAng);
//          printf("polAng=%6.2f\n",polAng);

        /* SMAP coordinates */

        sX = aziAng * sinf (polAng) / rad2deg + 180;
        sY = polAng / rad2deg;  /* + 1.5; */

        /* update */

        if (sX > 0 && sX < 360)
          if (sY > 0 && sY < 180)
            SMAP_firsthits->Fill (sX, sY, 1);
      };

  /* done */

  if (Pars.nprint > 0)
    fclose (fp);

  return (0);

}
